/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-baseXX.c,v 1.8 1999/03/31 17:59:54 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "common-stuff.h"

#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#define assert(x)   ((void*)0)
#endif

#include "peks-internal.h"
#include "peks-baseXX.h"
#include "baseXX.h"
#include "messages.h"
#include "crandom.h"
#include "cipher.h"
#include "rnd-pool.h"


/* ---------------------------------------------------------------------------- *
 *                 private functions: crc stuff                                 *
 * ---------------------------------------------------------------------------- */

static void
sequenced_md 
  (char result [4],
   const  char *x1,
   const  char *x2,
   const  char *x3)
{
  static frame_desc *frame ;
 
  if (frame == 0) {
    frame = create_frame 
      (find_frame_class (PEKS_KEY_LINE_CRC_TYPE), PEKS_KEY_LINE_CRC_OFFS);
    assert (frame != 0);
  }
  
  XCRCFIRST (frame, x1, strlen (x1)) ;
  XCRCNEXT  (frame, " ", 1) ;
  XCRCNEXT  (frame, x2, strlen (x2)) ;
  if (x3 != 0) {
    XCRCNEXT  (frame, " ", 1) ;
    XCRCNEXT  (frame, x3, strlen (x3)) ;
  }

  memcpy (result, XCRCRESULT (frame), 4) ;
}

/* ---------------------------------------------------------------------------- *
 *                  public functions: crc stuff                                 *
 * ---------------------------------------------------------------------------- */

char *
seqB64_md 
  (const  char *x1,
   const  char *x2,
   const  char *x3)
{
  unsigned char buf [4] ;
  sequenced_md (buf, x1, x2, x3) ;
  return bin2base64 (buf, 4);
}


int
comp_seqB64_md 
  (const  char *b64in,
   const  char *x1,
   const  char *x2,
   const  char *x3)
{
  char *t = seqB64_md (x1, x2, x3) ;
  int   n = strcmp (b64in, t) ;
  XFREE (t);
  return n;
}

/* ---------------------------------------------------------------------------- *
 *                     public mumeric conversion functions                      *
 * ---------------------------------------------------------------------------- */

char*
mpz2base64
  (mpz_t *OP)
{
  char *t, *s = ALLOCA (mpz_sizeinbase (*OP, 32) + 2) ;

  POINT_OF_RANDOM_STACK (6) ;

  mpz_get_str (s, 32, *OP) ;
  t = baseXtoBase64 (s,5) ;
  DEALLOCA (s) ;
  return t;
}


unsigned
base64toMpz
  (mpz_t     *OP,
   const char *s)
{
  char *t ;
  int n;
  if (s == 0) {
    mpz_set_ui (*OP, 0);
    return 0;
  }

  t = base64toBaseX (s,5) ;
  POINT_OF_RANDOM_VAR (t);

  n = mpz_set_str (*OP, t, 32);
  POINT_OF_RANDOM_VAR (n);

  POINT_OF_RANDOM_STACK (8) ;

  XFREE (t);
  return !n; /* 1 is ok, 0 is error */
}


char*
uint2base64
  (unsigned long n)
{
  char buf [2 * sizeof (long) + 1];

  POINT_OF_RANDOM_STACK (13) ;

  sprintf (buf, "%lx", n) ;
  return baseXtoBase64 (buf,4) ;
}


unsigned long
base64toUint
  (const char *s)
{
  unsigned long num ;
  char *t, *u, c ;

  POINT_OF_RANDOM_STACK (12) ;

  if (s == 0)
    return UINT_MAX ;

  t = base64toBaseX (s,4) ;
  POINT_OF_RANDOM_VAR (t);

  num = strtol (t, &u, 16) ;

  c = *u ; /* save char, u points into t */
  XFREE (t) ;
  return c ? UINT_MAX : num ; 
}
 
/* ---------------------------------------------------------------------------- *
 *                  bin to integer encapsulation                                *
 * ---------------------------------------------------------------------------- */

/*
 * encapsulated binary data text is repesented as the byte sequence
 *
 *	<len_high>, <len_low>, data bytes ...     padding ...
 *
 *                             <--- data text ---> 
 *      <-------------------------- buflen ------------------>
 *
 * and then converted to an mpz integer.
 */

int
mpzEncode
  (mpz_t        *OP,
   const char *text,
   unsigned     len,
   unsigned  buflen)
{
  char *t, *s;
  int n;

  if (len == 0)
    len = strlen (text) ;

  /* make sure that the arg text is not too long */
  if (len + 2 > buflen) {
    errno = errnum (ENC_TEXT_2LONG) ;
    return -1;
  }
  
  /* 
   * assemble text buffer: we set the high order bit so that we can 
   * make  sure about the length of the number when converted to an
   * mpz_t type number and back; leading zeros may savely be dropped
   * when converting back from the mpz_t type to the text
   */
  t = ALLOCA (buflen + 3) ;
  t [0] = ((len >> 8) & 0xff) | 0x80 ;
  t [1] =   len       & 0xff  ;
  memcpy (t + 2, text, len) ;

  POINT_OF_RANDOM_VAR (t);

  /* append some random noise if the text length is small */
  if (len + 2 < buflen)
    fast_random_bytes (t + 2 + len, buflen - len - 2) ;
  
  /* make base64 string */
  s = bin2base64 (t, len + 2) ;
  DEALLOCA (t);

  n = base64toMpz (OP, s) ;
  XFREE (s) ;

  if (n == 0) {
    errno = errnum (ENC_B64TEXT_ERR) ;
    return -1;
  }
  
  return 0 ;
}


char *
mpzDecodeB64
  (mpz_t *OP)
{
  unsigned n, len;
  char *t = mpz2base64 (OP) ;
  char *s = base64toBin (t, &n) ;
  
  XFREE (t);

  /* now, s = (<len_high>, <len_low>, bytes ...) */
  if (n < 2) {
    errno = errnum (ENC_B64EXTR_ERR) ;
    return 0;
  }
  
  len = ((s [0] & 0x7f) << 8) | s [1] ;
  if (len > n - 2) {
    errno = errnum (ENC_B64EXTR_ERR) ;
    return 0;
  }
  
  t = bin2base64 (s + 2, len) ;
  XFREE (s) ;
  return t;
}
