/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-client.c,v 1.14 1999/04/08 21:09:28 jordan Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#else
# include <socket.h> /* ????? */
#endif

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */

#include "peks-internal.h"
#include "peks-client.h"
#include "peks-setup.h"
#include "peks-handshake.h"
#include "peks-file.h"
#include "peks-string.h"
#include "crandom.h"
#include "messages.h"
#include "cbc-frame.h"
#include "baseXX.h"
#include "peks-baseXX.h"
#include "version.h"
#include "rnd-pool.h"


/* ---------------------------------------------------------------------------- *
 *                  private helpers                                             *
 * ---------------------------------------------------------------------------- */

static peks_key *
peks_client_setup
  (const char       *in,
   const char     *host,
   const char     *file,
   int abort_on_new_key)
{
  peks_key *key ;
  int n ;

  if ((key = accept_public_key_str (in)) == 0) 
    return 0;
  
  POINT_OF_RANDOM_STACK (7);

  /* check whether we know the server key, already */
  if (host != 0 && file != 0) {
    n = check_peks_sender_key (key, 0, host, 1 /*append missing entry*/, file) ;
    if (n < 0 || (n > 0 && abort_on_new_key)) {
      end_peks_key (key) ;
      return 0 ;
    }
  }

  POINT_OF_RANDOM_VAR (key);
  
  return key ;
}


static int
push_client_io_layer
  (unsigned        fd,
   const char   *what,
   const char    *key,
   unsigned       len,
   unsigned is_sender)

{
  void *cbc, *cc, *fc ;
  cipher_desc *cp ;
  frame_desc  *fp ;

  size_t c_size ;
  int  (*c_open) (void *, unsigned, cipher_desc*, frame_desc*, char[4]) ;
  int  (*c_rdwr) (void *, char *, unsigned, int);
  int  (*c_ioctl)(void *, io_ctrl_request, void*);
  void (*c_close)(void *) ;

  /* get new cipher and frame classes to be used, here */
  if (find_classes_by_name (&cc, &fc, what) == 0)
    return -1;

  /* create an cipher instance and a frame */
  if ((cp = (*(is_sender?create_encryption:create_decryption)) (cc, key, len)) < 0)
    return -1;
  fp = create_frame (fc, key [len ? len-1 : 0]) ;
  
  /* push that new io layer */
  cbc_get_info (is_sender, &c_size, &c_open, &c_rdwr, &c_ioctl, &c_close);
  if ((cbc = io_push (fd, c_size, c_rdwr, c_ioctl, c_close, is_sender)) == 0) {
    destroy_cipher (cp);
    destroy_frame  (fp);
    return -1;
  }
  
  /* initialize that io layer */
  return (*c_open) (cbc, fd, cp, fp, (char*)key + (len<5 ? 0 : len-5)) ;
}

/* ---------------------------------------------------------------------------- *
 *                  public key client handshake function                        *
 * ---------------------------------------------------------------------------- */

int
client_negotiate_session_key
  (const char  * cipher,
   int           socket,
   const char  * server,
   const char * keyfile)
{
  peks_key *key ;
  char *s, *u, *buf ;
  int n, l ;

  /* initialize prime random generator */
  init_random_gen ((char*)&s, sizeof (s)) ;

  /* read the public key from the server host */
  buf = ALLOCA (CBC_IO_BUFLEN_MAX+1) ;
  if ((n = OS_RECV (socket, buf, CBC_IO_BUFLEN_MAX, 0)) < 0) {
    DEALLOCA (buf) ;
    return -1 ;
  }
  buf [n] = 0 ;

  POINT_OF_RANDOM_VAR (buf);

  /* arg 4 set means: return an error if a new key has been stored */
  if ((key = peks_client_setup (buf, server, keyfile, 1)) == 0) {
    /* send a termination message to the server */
    OS_SEND (socket, ".", 2, 0) ;
    goto error_code_return;
  }

  POINT_OF_RANDOM_STACK (13);
  
  /* generate the receiver session key */
  if ((l = cipher_keylen (cipher) + 5) == 5) {
    end_peks_key (key) ;
    goto error_code_return;
  }
  prime_random_bytes (s = ALLOCA (l), l) ;
  
  /* push a new io layer for reading */
  if (push_client_io_layer (socket, cipher, s, l, 0) < 0) {
    end_peks_key (key) ;
    DEALLOCA (s);
    goto error_code_return;
  }

  POINT_OF_RANDOM_VAR (s);
  
  /*  wrap the client receiver session key and send it */
  u = peks_wrap_session_key (s, l, cipher) ;
  DEALLOCA (s);
  s = make_elg_response_key_str (key, u, strlen (u)) ;
  XFREE (u);
  end_peks_key (key) ;			/* not needed, any longer */
  if (s == 0) 
    goto error_code_return;
  OS_SEND (socket, s, strlen (s), 0);
  POINT_OF_RANDOM_VAR (s);
  XFREE (s) ;
  
  /* receive the sender session key s from the server */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0)
    goto error_code_return;
  buf [n] = '\0' ;
  if ((s = peks_unwrap_session_key ((const char **)&u, &l, buf)) == 0)
    goto error_code_return;

  POINT_OF_RANDOM_VAR (s);
    
  /* push a new io layer for sending */
  n = push_client_io_layer (socket, u, s, l, 1) ;
  XFREE (s);
  XFREE (u);
  if (n != 0)
    goto error_code_return;
  
  /* send a magig string in order to test the connection */
  if (io_send (socket, PEKS_MAGIC, strlen (PEKS_MAGIC)+1, 0) !=
      strlen (PEKS_MAGIC)+1) {
    errno = errnum (CLNT_MAGIC_FAILED) ;
    goto error_code_return;
  }

  POINT_OF_RANDOM_VAR (buf);
  DEALLOCA (buf) ;
  return 0;
  /*@NOTREACHED@*/

 error_code_return:
  io_pop (socket, 2) ; /* its's save to always execute it */
  DEALLOCA (buf) ;
  return -1 ;
}

/* ---------------------------------------------------------------------------- *
 *                  public client authentication function                       *
 * ---------------------------------------------------------------------------- */


int 
peks_client_authenticate
  (unsigned    type,
   int       socket, 
   const char *user, 
   peks_key    *key, 
   const char *(*get_pwd) (void))
{
  const char *pwd;
  char buf [CBC_IO_BUFLEN_MAX + 1];
  char *s, *md = PEKS_KEY_LINE_CRC_TYPE ;
  int n, r_code = -1 ;

  POINT_OF_RANDOM_STACK (14);

  /* initialize prime random generator */
  init_random_gen ((char*)&s, sizeof (s)) ;

  if (key == 0) {
    errno = errnum (FUNCTION_ARG_ERROR);
    return -1;
  }
  
  if (key->pubkey_str == 0) 
    key->pubkey_str = make_pubkey_line (key) ;
  
  /* send the public peks key to the server */
  n = strlen (key->pubkey_str) ;
  if (io_send (socket, key->pubkey_str, n + 1, 0) < n) 
    goto error_code_return ;

  POINT_OF_RANDOM_VAR (buf);

  /* accept the server challenge */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN_MAX, 0)) <= 2) {
    /* check for a termination message from the server */
    if (n == 2 && buf [0] == '.' && buf [1] == '\0')
      errno = errnum (CLNT_AUTH_TERM);
    else if (n >= 0)
      errno = errnum (CLNT_AUTH_RESPERR);
    goto error_code_return ;
  }
  if (strncmp (buf, "fail", 3) == 0)	/* login rejected */
    goto ok_return_code ;
  buf [n] = '\0' ;
  if (accept_challenge_str (key, buf) < 0) 
    goto error_code_return ;

  /* return the signed challenge to the server */
  if ((s = sign_challenge_str (type, user, key)) == 0)
    goto error_code_return ;
  n = io_send (socket, s, strlen (s) + 1, 0) ;
  XFREE (s) ;
  if (n < 0)
    goto error_code_return ;

  /* get the server response */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN_MAX, 0)) < 0) 
    errno = errnum (CLNT_AUTH_RESPERR);
  if (strncmp (buf, "ok", 2) == 0)	/* ok => login accepted */
    goto ok_return_code ;
  if (strncmp (buf, "next", 5) != 0) {	/* not retry => login rejected */
    r_code = 1 ;
    goto error_code_return ;
  }
  
 
  if (get_pwd == 0 || (pwd = (*get_pwd) ()) == 0) {
    io_send (socket, ".", 2, 0) ;
    goto error_code_return ;
  }

  /* make passwd digest and send it */
  if ((s = make_digest_challenge_str (key, md, user, pwd)) == 0)
    goto error_code_return ;
  if (io_send (socket, s, strlen (s) + 1, 0) < 0)
    goto error_code_return ;

  /* get the server response */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN_MAX, 0)) <= 0) 
    goto error_code_return ;
 
  if (strncmp (buf, "ok", 2) == 0)	/* ok => login accepted */
    goto ok_return_code ;

  /* otherwise => login rejected */
  r_code = 1 ;
  goto error_code_return ;
  
 ok_return_code:
  r_code = 0 ;

 error_code_return:
  end_peks_key_session (key) ;
  return r_code;
}
