/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *   Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-handshake.c,v 1.10 1999/04/06 20:08:44 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#include "peks-internal.h"
#include "peks-handshake.h"
#include "peks-string.h"
#include "crandom.h"
#include "elgamal.h"
#include "make-primes.h"
#include "peks-baseXX.h"
#include "messages.h"
#include "baseXX.h"
#include "version.h"
#include "rnd-pool.h"

/* ---------------------------------------------------------------------------- *
 *                  public key handshake functions                              *
 * ---------------------------------------------------------------------------- */

/*
 * Accept a public key
 *
 *      "peks/x.y:" <p> <g> <g^a> <crc>
 *
 * and return an initialized client key.  This is done in two
 * steps to allow a key check against a host key data base, in
 * beween .
 */

peks_key *
accept_public_key_str
  (const char *line)
{
  const char *s, *proto [] = {"peks", 0} ;
  unsigned version = peks_split_ident (proto, line, 0) ;

  POINT_OF_RANDOM_VAR (version);

  /* skip over version string */
  if ((s = strchr (line, ':')) == 0 || !(++ s, isspace (*s))) {
    errno = errnum (PEKS_VERS_MISMATCH) ;
    return 0;
  }

  /* check for version compatibility */
  switch (version) {
  case (100 + PEKSMAJOR) * 100 + PEKSMINOR:

    POINT_OF_RANDOM_STACK (11);

    /*
     * modulus   == p
     * generator == g
     * private   == g^a (mod p)
     */
    
    return peks_parse_encryption_line (++ s) ;
  }

  errno = errnum (PEKS_VERS_MISMATCH) ;
  return 0;
}


/*
 * make a response key string after accepting the public key
 *
 *	"elg/x.y:" <g^b> <text * g^ab> <crc>
 *
 * ready to complete the El Gamal key negotiation;
 *
 */

char *
make_elg_response_key_str
  (peks_key      *k,
   const char *text,
   unsigned     len)
{
  char *s, *t ;

  POINT_OF_RANDOM_STACK (11) ;

  if ((t = el_gamal_encrypt_line (&k->modulus,	/* p */
				  k->generator,	/* g */
				  &k->private,	/* g^a */
				  text, len)) == 0)
    return 0 ;

  /* make the response line */
  sprintf (s = XMALLOC (sizeof (ELG_VERSION)+2+strlen (t)+1), "%s: %s", ELG_VERSION, t);
  XFREE (t) ;

  POINT_OF_RANDOM_VAR (t) ;
  return s ;
}


/*
 * accept a response key 
 *
 *      xx/#.#: <pub_key> "A" <value> <crc>
 *
 * as generated with make_xx_response_key_str () and update 
 * the server key structure
 */

int
accept_response_key_str
  (peks_key    *key,
   const char *line)
{
  const char *s, *proto [] = {"elg", 0};
  unsigned         version = peks_split_ident (proto, line, 0) ;
  
  if ((s = strchr (line, ':')) == 0 || !(++ s, isspace (*s))) {
    errno = errnum (PEKS_VERS_MISMATCH) ;
    return -1;
  }

  POINT_OF_RANDOM_VAR (version);

  /* check for version compatibility */
  switch (version) {

  case (100 + PEKSMAJOR) * 100 + PEKSMINOR: /* El Gamal */
  
    if ((key->import_str = (char *)
	 b64_el_gamal_decrypt_line (&key->modulus, &key->private, ++ s)) != 0) {
      POINT_OF_RANDOM_VAR (line);
      return 0;
    }
  }

  errno = errnum (PEKS_VERS_MISMATCH) ;
  return -1;
}


/* ---------------------------------------------------------------------------- *
 *                public signature key handshake functions                      *
 * ---------------------------------------------------------------------------- */

/*
 * make a challenge string after accepting the public key
 *
 *	"chl: " <random-text>
 *
 * where the random text is base 64 encoded
 */

char *
make_challenge_str		/* does not return NULL */
  (peks_key *k)
{
  char *s = ALLOCA (AUTH_CHALLENGE_LEN) ;

  prime_random_bytes (s, AUTH_CHALLENGE_LEN) ;
  k->challg_str = bin2base64 (s, AUTH_CHALLENGE_LEN) ;
  DEALLOCA (s) ;

  sprintf (s = XMALLOC (strlen (k->challg_str) + 6), "chl: %s", k->challg_str);
  return s ;
}  


int
accept_challenge_str		/* does not return NULL */
  (peks_key  *key, 
   const char *in)
{
  if (strncmp ("chl:", in, 4) != 0 || !(in += 4, isspace (*in))) {
    errno = errnum (CLNT_AUTH_RESPERR);
    return -1;
  }
  in ++ ;

#if 0  
  /* strlen (in) == (4 * raw-data-lenghth + 2) / 3, so we won't tolerate 
     challenges with  3 * len < 4 * raw-data-lenghth + 2 */
  if (((mpz_sizeinbase (key->modulus, 2) - 1) >> 3) < strlen (in)) {
    errno = errnum (CLNT_AUTH_SHORTCHALL) ;
    return -1;
  }
#endif

  if (key->challg_str != 0)
    XFREE (key->challg_str) ;
  key->challg_str = XSTRDUP (in);

  return 0 ;
}  


/*
 * send/accept user name and challenge signature
 *
 *	"egs/n.m: " <user> <seed> <sig msg> <crc>
 *
 * where the <> fields are base 64 encoded
 */

char *
sign_challenge_str
  (int        type,
   const char *usr,
   peks_key     *k)
{
  char *t, *s ;

  if ((s = el_gamal_sign_line 
       (type, usr, k->challg_str, &k->modulus, k->generator, &k->private)) == 0)
    return 0 ;

  sprintf (t = XMALLOC (strlen (s) + 25), "egs/%u.%u: %s", type, 0, s) ;
  XFREE (s);
  return t ;
}

char *
accept_signed_challenge_str
  (peks_key      *k,
   const char *line)
{
  const char *s, *proto [] = {"egs", 0};
  unsigned       variation = peks_split_ident (proto, line, 0) ;
  
  /* check for the signature variation as listed in table 11.5 from
     A. Menezes, P.van Oorschot, S. Vanstone (Eds.), "Handbook of Applied 
     Cryptography", CRC Press 1997, ISBN 0-8493-8523-7, pg 457 ff. */
  
  switch (variation) {
  case (100 + 1) * 100 + 0: /* sig: (m - a (g^b)) / b (mod p-1), b = random */
    variation = 1;
    break;
  case (100 + 3) * 100 + 0: /* sig: m b + a (g^b) (mod p-1), b = random */
    variation = 3 ;
    break;
  default:
    errno = errnum (PEKS_VERS_MISMATCH) ;
    return 0;
  }

  POINT_OF_RANDOM_VAR (variation);

  /* skip over version string */
  if ((s = strchr (line, ':')) == 0 || !(++ s, isspace (*s))) {
    errno = errnum (PEKS_VERS_MISMATCH) ;
    return 0;
  }
  
  return el_gamal_verify_line
    (variation,
     k->challg_str,	/* message */
     &k->modulus,	/* prime p */
     k->generator,	/* generator g (mod p) */
     &k->private,	/* public key g^a */
     ++ s);
}


/*
 * send/accept user name and challenge signature
 *
 *	"md5: " <digest (challenge+passwd)>
 *
 * where the <digest> field is  base 64 encoded
 */

char *
make_digest_challenge_str
  (peks_key    *key,
   const char  *tag,
   const char  *usr,
   const char  *pwd)
{
  char *s, *t;

  /* make the string: <user> <passwd> */
  sprintf (s = ALLOCA (strlen (usr) + 1 + strlen (pwd) + 1), 
	   "%s %s", usr, pwd);

  t = peks_digest (tag, key->challg_str, 0, s, 0) ;
  DEALLOCA (s);
  return t;
}


int
accept_digest_challenge_str
  (const peks_key *key,
   const char  *challg,
   const char     *pwd,
   const char    *line)
{
  char *s, *tag;
  int n ;
  
  if (key == 0 || challg == 0 ||
      (s = strchr (line, ':')) == 0 || !(++ s, isspace (*s))) {
    errno = errnum (ILL_DIGST_HEADER);
    return -1;
  }
  
  /* get message digest as fiven in the header */
  n = (s - line) -1 ;
  memcpy (tag = ALLOCA (n + 1), line, n);
  tag [n] = '\0' ;

  s = peks_digest (tag, challg, 0, pwd, 0) ;
  DEALLOCA (tag);

  /* compare */
  n = strcmp (s, line) ;
  XFREE (s) ;

  return n ;
}
