/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *       Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-server.c,v 1.10 1999/03/31 17:59:56 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#else
# include <socket.h> /* ????? */
#endif

#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif

#define __IOSTREAM_PRIVATE__ /* prevent from using send/recv auto-replaces */

#include "peks-internal.h"
#include "peks-server.h"
#include "peks-setup.h"
#include "peks-handshake.h"
#include "peks-file.h"
#include "peks-string.h"
#include "crandom.h"
#include "messages.h"
#include "cbc-frame.h"
#include "baseXX.h"
#include "version.h"
#include "rnd-pool.h"
#include "hostnames.h"

/* ---------------------------------------------------------------------------- *
 *                  private key handshake functions                             *
 * ---------------------------------------------------------------------------- */


static char *
peks_server_decode_sender_key
  (peks_key    *k,
   unsigned    *N,
   const char **T,
   const char *in)
{
  char *s, *sk ;
  unsigned dummy ;

  if (N == 0)
    N = &dummy ;

  POINT_OF_RANDOM_STACK (7);

  /* in case of an error, errno will be set */
  if (accept_response_key_str (k, in) < 0)
    return 0 ;
  
  /* extract session key s of length n */
  s  = base64toBin (k->import_str, N) ;
  sk = peks_unwrap_session_key (T, N, s) ;
  XFREE (s);
  return sk;
}




static int
push_server_io_layer
  (unsigned        fd,
   const char   *what,
   const char    *key,
   unsigned       len,
   unsigned is_sender)

{
  void *cbc, *cc, *fc ;
  cipher_desc *cp ;
  frame_desc  *fp ;

  size_t c_size ;
  int  (*c_open) (void *, unsigned, cipher_desc*, frame_desc*, char[4]) ;
  int  (*c_rdwr) (void *, char *, unsigned, int);
  int  (*c_ioctl)(void *, io_ctrl_request, void*);
  void (*c_close)(void *) ;

  /* get new cipher and frame classes to be used, here */
  if (find_classes_by_name (&cc, &fc, what) == 0)
    return -1;

  /* create an cipher instance and a frame */
  if ((cp = (*(is_sender?create_encryption:create_decryption)) (cc, key, len)) < 0)
    return -1;
  fp = create_frame (fc, key [len ? len-1 : 0]) ;
  
  /* push that new io layer */
  cbc_get_info (is_sender, &c_size, &c_open, &c_rdwr, &c_ioctl, &c_close);
  if ((cbc = io_push (fd, c_size, c_rdwr, c_ioctl, c_close, is_sender)) == 0) {
    destroy_cipher (cp);
    destroy_frame  (fp);
    return -1;
  }

  POINT_OF_RANDOM_STACK (11);
  
  /* initialize that io layer */
  return (*c_open) (cbc, fd, cp, fp, (char*)key + (len<5 ? 0 : len-5)) ;
}


/* ---------------------------------------------------------------------------- *
 *                  public key server auth function                             *
 * ---------------------------------------------------------------------------- */

static char *
peks_authenticate
  (peks_key   *pwkey,
   char       *iobuf,
   unsigned   buflen,
   unsigned   keylen,
   unsigned max_pwd_failures,
   const char *kfile)
{
  char *pwd = 0, *usr = 0, *pwusr = 0, *r_code = 0;
 
  char *s ;
  peks_key *key ;
  int n ;

  if ((key = accept_public_key_str (iobuf)) == 0) 
    goto reject_return ;

  if ((n = peks_keylen (key)) < keylen) {
    char msg [100] ; 
    sprintf (msg, peks_strerr (errnum (PUBKEY_2SHORT_UU)), n, keylen);
    errno = peks_errnum (msg) ;
    goto reject_return ;
  }

  /* send challenge string */
  s = make_challenge_str (key);
  n = io_send (pwkey->socket, s, strlen (s), 0) ;
  XFREE (s);
  if (n < 0) 
    goto reject_return ;

  /* receive signature -> usr */
  if ((n = io_recv (pwkey->socket, iobuf, buflen - 1, 0)) < 0) 
    goto error_return ;
  iobuf [n] = '\0' ;
  if ((usr = accept_signed_challenge_str (key, iobuf)) == 0) 
    goto reject_return ;
  
  /* check whether we know that user, already */
  if ((n = check_peks_sender_key 
       (key, usr, pwkey->host_str, 0 /*don't create*/, kfile)) < 0) 
    goto reject_return ;
  if (n == 0) 
    goto accept_return ;
  
  /* do we have a second chance ? */
  if (kfile == 0) 
    goto reject_return ;
  
  /* the user does not exist, yet - check the passwd data base */
  if (pwkey->host_str != 0) {
    pwusr = ALLOCA (strlen (usr) + 1 + strlen (pwkey->host_str) + 1) ;
    sprintf (pwusr, "%s@%s", usr, pwkey->host_str) ;
    pwd = peks_edit_passwd_file (pwkey, pwusr, 0, 0, kfile) ;
  }
  if (pwd == 0) {
    if (pwusr != 0)
      DEALLOCA (pwusr) ;
    pwusr = ALLOCASTRDUP (usr) ;
    if ((pwd = peks_edit_passwd_file (pwkey, pwusr, 0, 0, kfile)) == 0) 
      goto reject_return ;
  }

  /* good, so let the user send a password digest */
  if (io_send (pwkey->socket, "next", 5, 0) < 0) 
    goto error_return ;
  
  /* receive and evaluate that digest */
  if ((n = io_recv (pwkey->socket, iobuf, buflen - 1, 0)) < 0 || 
      /* check for a termination message from the client */
      (n > 0 && iobuf [0] == '.')) {
    end_peks_key_session (pwkey) ;
    goto error_return ;
  }

  /* the pwd entry looks like: <0 + # failures> <user name> <passwd>
     so we need to skip the first entry */
  if (pwd [0] == 0 || !isspace (pwd [1])) {
    /* corrupt entry - remove the passwd */
    peks_edit_passwd_file (0, pwusr, 0, 0, kfile) ;
    goto reject_return ;
  }

  if (accept_digest_challenge_str (pwkey, key->challg_str, pwd+2, iobuf) != 0) {
    /* check the number of unsuccessful logins */
    if (pwd [0] >= '0' + max_pwd_failures || pwd [0] >= '0' + MAX_PWD_FAILURES)
      /* remove the passwd */
      peks_edit_passwd_file (0, pwusr, 0, 0, kfile) ;
    else {
      /* edit that password entry, store incremented failure count */
      if ((s = strchr (pwd + 2, ' ')) != 0) s++ ;
      peks_edit_passwd_file (pwkey, pwusr, pwd [0] - '0' + 1, s, kfile) ;
    }
    goto reject_return ;
  }

  /* remove the passwd and store that pub key in the data base */
  if (peks_edit_passwd_file (0, pwusr, 0, 0, kfile) == 0 ||
      create_peks_keyfile (key, usr, pwkey->host_str, kfile) < 0) {
    peks_server_auth_reject (pwkey, peks_strerr (errno));
    goto error_return ;
  }
  
 accept_return:
  r_code = usr ;  /* accept */
  usr    =   0 ;

 error_return:
  end_peks_key (key) ;

  if (pwusr != 0) DEALLOCA (pwusr);
  if (usr   != 0)    XFREE   (usr);
  if (pwd   != 0)    XFREE   (pwd);

  if (r_code == 0) /* reset session record */
    end_peks_key_session (pwkey) ;

  return r_code;
  /*@NOTREACHED@*/

 reject_return:
  peks_server_auth_reject (pwkey, 0);
  goto error_return ;
} 


/* ---------------------------------------------------------------------------- *
 *                  public key server handshake function                        *
 * ---------------------------------------------------------------------------- */

int
server_negotiate_session_key
  (peks_key       *defk,
   int           socket,
   const char  * client,
   const char * keyfile)
{
  const char *cipher, *t ;
  peks_key *key ;
  char *s, *buf;
  int l, n;

  POINT_OF_RANDOM_STACK (14);

  /* initialize prime random generator */
  init_random_gen ((char*)&s, sizeof (s)) ;

  /* set up the public peks server key:
   *
   * 1. search for a key in the keyfile, first
   * 2. use the default key, next
   * 3. generate a key finally 
   */

  /* if ((t = peks_server_setup (&key, client, keyfile, key)) == 0) { */

  key = client != 0 && keyfile != 0 
    ? read_peks_key (0, client, -1, keyfile) 
    : 0
    ;

  POINT_OF_RANDOM_STACK (13);

  if (key == 0)
    key = (defk != 0 ? dup_peks_key (defk) : new_peks_key (0)) ;

  if (key == 0) {
    errno = errnum (SRV_SETUP_SKEY) ;
    return -1;
  }
  
  if (key->pubkey_str == 0) 
    key->pubkey_str = make_pubkey_line (key) ;
  
  /* send the public peks key to the client */
  OS_SEND (socket, key->pubkey_str, strlen (key->pubkey_str) + 1, 0);
  POINT_OF_RANDOM_VAR (t);

  /* receive the peks wrapped answer from the client */
  buf = ALLOCA (CBC_IO_BUFLEN_MAX+1) ;
  if ((n = OS_RECV (socket, buf, CBC_IO_BUFLEN_MAX, 0)) < 0)
    goto error_code_return ;
  /* check for a termination message from the client */
  if (n > 0 && buf [0] == '.') {
    errno = errnum (SRV_CLIENT_KNOCKING);
    goto error_code_return ;
  }
  buf [n] = '\0' ;

  /* unpack the session key from the peks wrapper */
  s = peks_server_decode_sender_key (key, &l, &cipher, buf) ;
  end_peks_key (key) ;
  if (s == 0)
    goto error_code_return ;
 
  /* install it as sender session key s of length l and type t */
  n = push_server_io_layer (socket, cipher, s, l, 1) ;
  POINT_OF_RANDOM_VAR (s);
  XFREE (s);
  if (n < 0)
    goto error_code_return ;
  
  /* generate the receiver session key */
  if ((l = cipher_keylen (cipher) + 5) == 5)
    goto error_code_return;
  prime_random_bytes (s = ALLOCA (l), l) ;
   
  /* push a new io layer for reading */
  if (push_server_io_layer (socket, cipher, s, l, 0) < 0) {
    DEALLOCA (s);
    goto error_code_return ;
  }

  /* peks_wrap_session_key () always returns a value, send it */
  buf = peks_wrap_session_key (s, l, cipher) ;
  POINT_OF_RANDOM_VAR (cipher);
  XFREE (cipher);
  POINT_OF_RANDOM_VAR (s);
  DEALLOCA (s);
  if (io_send (socket, buf, strlen (buf), 0) < 0)
    goto error_code_return ;
  
  /* check magic string */
  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0) {
    errno = errnum (SRV_MAGIC_FAILED) ;
    goto error_code_return ;
  }
  buf [n] = '\0' ;
  if (strcmp (buf, PEKS_MAGIC) != 0) {
    errno = errnum (SRV_WRONG_MAGIC) ;
    goto error_code_return ;
  }

  /* done */
  DEALLOCA (buf) ;
  return 0;
  /*@NOTREACHED@*/ 

 error_code_return:
  io_pop (socket, 2) ; /* its's always save to execute it */
  DEALLOCA (buf) ;
  return -1 ;
}

/* ---------------------------------------------------------------------------- *
 *                  server authentication function                              *
 * ---------------------------------------------------------------------------- */

char *
peks_server_auth_pubkey
  (int       socket,
   const char *host,
   peks_key  *pwkey,
   unsigned  keylen,
   unsigned max_pwd_failures,
   const char *keyf)
{
  char *buf, *s;
  unsigned n, version ;

  /* this is the ordered list of authentication schemes, available */
  const char *proto [] = {"peks", 0};

  /* initialize prime random generator */
  init_random_gen ((char*)&s, sizeof (s)) ;

  POINT_OF_RANDOM_STACK (7);

  if (host == 0 || pwkey == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return 0 ;
  }

  /* receive the auth request from the client */
  buf = ALLOCA (CBC_IO_BUFLEN_MAX+1) ;

  if ((n = io_recv (socket, buf, CBC_IO_BUFLEN, 0)) < 0)
    goto error_code_return ;
  buf [n] = '\0' ;
  if ((s = strchr (buf, ':')) == 0 || !(++s, isspace (*s))) {
    errno = errnum (AUTH_HEADER_FAILED) ;
    goto error_code_return ;
  }

  POINT_OF_RANDOM_VAR (buf);

  /* check for the authentication scheme, to be used */
  if ((version = peks_split_ident (proto, buf, s - buf)) == 0) {
    errno = errnum (AUTH_HEADER_FAILED) ;
    goto error_code_return ;
  }

  POINT_OF_RANDOM_VAR (version);
  
  pwkey->host_str = XSTRDUP (host) ;
  pwkey->socket   = socket ;

  if (10000 <= version && version < 20000) {

    if (version == (100 + PEKSMAJOR) * 100 + PEKSMINOR) {
      
      /* pass the working buffer with the first line */
      s = peks_authenticate (pwkey, buf, CBC_IO_BUFLEN_MAX+1, 
			     keylen, max_pwd_failures, keyf);
      DEALLOCA (buf) ;
      return s ; /* return the user name */
    }
    errno = errnum (AUTH_HEADER_FAILED) ; 
    goto error_code_return ;
  }

  /* you might proceed to other authentication schemes */
  pwkey->import_str = bin2base64 (buf, n) ;
  errno = 0 ;
  
 error_code_return:
  DEALLOCA (buf) ;
  if (pwkey->host_str != 0) {
    XFREE (pwkey->host_str) ;
    pwkey->host_str = 0 ;
  }
  return 0 ;
}


void
peks_server_auth_accept
  (peks_key *key)
{
  if (key->socket >= 0) 
    io_send (key->socket, "ok", 3, 0) ;
  end_peks_key_session (key) ;
}


void
peks_server_auth_reject
  (peks_key   *key,
   const char *info)
{
  if (key->socket >= 0) {
    char *s ;
  
    if (info == 0) 
      s = "failed" ;
    else
      sprintf (s = ALLOCA (3 + strlen (info) + 1), "failed %s", info) ;
    
    io_send (key->socket, s, strlen (s) + 1, 0) ;

    if (info != 0)
      DEALLOCA (s) ;
  }

  end_peks_key_session (key) ;
}
