/*
 *          Copyright (c) mjh-EDV Beratung, 1996-1999
 *     mjh-EDV Beratung - 63263 Neu-Isenburg - Rosenstrasse 12
 *          Tel +49 6102 328279 - Fax +49 6102 328278
 *                Email info@mjh.teddy-net.com
 *
 *    Author: Jordan Hrycaj <jordan@mjh.teddy-net.com>
 *
 *   $Id: peks-string.c,v 1.2 1999/03/31 17:59:56 renaud Exp $
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   PEKS - private exponent key stuff for Diffie Hellman and El Gamal
 */

#include "common-stuff.h"
#include "peks-internal.h"
#include "peks-string.h"
#include "peks-setup.h"
#include "peks-baseXX.h"
#include "baseXX.h"
#include "cipher.h"
#include "messages.h"
#include "rnd-pool.h"
#include "version.h"

/* ---------------------------------------------------------------------------- *
 *                  public functions: line encode                               *
 * ---------------------------------------------------------------------------- */

/*
 * For a given key structure, return a text
 *
 *    <X> <Y> <Z> <crc>
 *
 * where CRC is an hash over the preceeding text fields.
 */

char *
b64_make_encryption_line 
  (const mpz_t    *X,
   const unsigned *Y,
   const mpz_t    *Z)
{
  char *x = 0, *y = 0, *z = 0, *crc, *s;
  unsigned len = 0;

  if (X != 0) {x =  mpz2base64 ((mpz_t*)X); len += strlen (x) + 1; }
  if (Y != 0) {y = uint2base64        (*Y); len += strlen (y) + 1; }
  if (Z != 0) {z =  mpz2base64 ((mpz_t*)Z); len += strlen (z) + 1; }

  crc = seqB64_md (x, y, z);

  /* make the response line */
  s = XMALLOC (len + strlen (crc) + 1);

  s [0] = '\0' ;
  if (x != 0) { strcat (s, x); strcat (s, " "); XFREE (x); }
  if (y != 0) { strcat (s, y); strcat (s, " "); XFREE (y); }
  if (z != 0) { strcat (s, z); strcat (s, " "); XFREE (z); }
  strcat (s, crc) ;
 
  XFREE (crc);
  return s;
}


/*
 * For a given key structure, return a text
 *
 *   <tag>: <modulus> <generator> <private> <crc>
 *
 * where CRC is an hash over the preceeding three text fields.
 */

char *
make_peks_key_line
  (const char     *usr,
   const char    *host,
   const peks_key *key)
{
  char *line = b64_make_encryption_line 
    (&key->modulus, &key->generator, &key->private) ;

  char *tag  = ALLOCA ((usr  == 0 ? 0 : strlen (usr)) + 
		       (host == 0 ? 0 : strlen (host)) + 3) ;
  char *s ;

  tag [0] = '\0' ;
  if (usr != 0) {
    strcat (tag, usr) ;
    strcat (tag, "@") ;
  }
  if (host != 0) {
    strcat (tag, host) ;
    strcat (tag, ":") ;
  }

  sprintf (s = XMALLOC (strlen (tag)+1+strlen (line)+1), "%s %s", tag, line) ;
			
  DEALLOCA (tag);
  XFREE   (line);
  return s ;
}


/*
 * make a public server key string ready for export
 *
 *      "peks/x.y:" <p> <g> <g^a> <crc>
 */

char *
make_pubkey_line
  (const peks_key *k)
{
  peks_key *dup = dup_peks_key (k);
  
  mpz_t G ;
  char *s ;

  POINT_OF_RANDOM_STACK (5);

  mpz_init_set_ui (G, k->generator) ;
  mpz_powm (dup->private, G, k->private, k->modulus);	/* pub key := g^a (mod p) */
  mpz_clear (G) ;

  s = make_peks_key_line (0, PEKS_VERSION, dup) ;
  end_peks_key (dup) ;
  return s;
}


/* ---------------------------------------------------------------------------- *
 *                  public functions: line decode                               *
 * ---------------------------------------------------------------------------- */

/*
 * For a given list of strings [] = {<ident>, ... } and two non negative
 * numbers <major> < 100, <minor> < 100 parse a string 
 *
 *	<ident> "/" <major> "." <minor>
 *
 * and return the unique identifier
 *
 *	(index {<ident>} + 1) * 10000 + <major> * 100	+ <minor>
 *
 * if such an index exists, and 0 otherwise.
 */

unsigned
peks_split_ident
  (const char *ident_table [], /* NULL terminated array */
   const char           *text,
   unsigned               len)
{
  char *t, *s ;
  int maj, min, n = 0, result = 0;

  POINT_OF_RANDOM_STACK (14);

  /* get a writable copy of the argument string */
  if (len == 0) 
    len = strlen (text) ;
  (s = strncpy (ALLOCA (len + 1), text, len + 1)) [len] = '\0' ;

  if ((t = strtok (s, "/")) != 0)
    
    while (ident_table [n] != 0) {
      
      /* get a match in the lookup table */
      if (strcmp (ident_table [n ++], t) != 0)
	continue ;

      /* ok, get the rest */
      if ((t = strtok (0, PEKS_FS)) != 0 &&
	  sscanf (t, "%u.%u", &maj, &min) == 2 &&
	  maj < 100 && min < 100)
	result = (n * 100 + maj) * 100 + min ;
      break;
    }
  
  DEALLOCA (s) ;
  return result ;
}


int
parse_encryption_line
  (char **X,
   char **Y,
   char **Z,
   const char *in)
{
  char *x = 0, *y = 0, *z = 0, *crc;
  
  /* dup argument string: we need to edit */
  char *t = ALLOCASTRDUP (in) ;

  if (X != 0 && (x = strtok (t, PEKS_FS)) == 0) 
    goto syntax_error;
  if (Y != 0 && (y = strtok (0, PEKS_FS)) == 0) 
    goto syntax_error;
  if (Z != 0 && (z = strtok (0, PEKS_FS)) == 0) 
    goto syntax_error;
  if ((crc = strtok (0, PEKS_FS)) == 0) 
    goto syntax_error;
  if (strtok (0, PEKS_FS) != 0) 
    goto syntax_error;
  
  /* verify line entries */
  if (comp_seqB64_md (crc, x, y, z) != 0) {
    errno = errnum (LINE_CRC_ERR) ;
    DEALLOCA (t);
    return -1;
  }
  
  if (X != 0) *X = XSTRDUP (x) ;
  if (Y != 0) *Y = XSTRDUP (y) ;
  if (Z != 0) *Z = XSTRDUP (z) ;
  return 0;

 syntax_error:
  errno = errnum (LINE_SYNTAX_ERR) ;
  DEALLOCA (t);
  return -1;
}


peks_key *
peks_parse_encryption_line
  (const char *in)
{
  peks_key *key ;
  char *x, *y, *z;
  
  if (parse_encryption_line (&x, &y, &z, in) < 0)
    return 0;

  /* create key structure */
  key = XMALLOC (sizeof (peks_key));
 
  /* initialize */
  key->socket = -1 ;

  /* convert to internal number representation */
  mpz_init (key->modulus) ;
  mpz_init (key->private) ;
  
  POINT_OF_RANDOM_VAR (x);

  if (base64toMpz (&key->modulus, x) == 0) 
    goto error_return;
  if ((key->generator = base64toUint (y)) == UINT_MAX) 
    goto error_return;
  if (base64toMpz (&key->private, z) == 0) 
    goto error_return;
  
  if (peks_keylen (key) < MINIMUM_PRIME_BITS) {
    errno = errnum (LINE_KEYLEN_UNDERFLW);
    goto x_error_return;
  }

  POINT_OF_RANDOM_STACK (9);

  XFREE (x) ;
  XFREE (y) ;
  XFREE (z) ;

  return key;
  /*@NOTREACHED@*/

 error_return:

  errno = errnum (LINE_PARSER_B64ERR);

 x_error_return:
  
  mpz_clear (key->modulus) ;
  mpz_clear (key->private) ;
  
  XFREE (key) ;
  XFREE   (x) ;
  XFREE   (y) ;
  XFREE   (z) ;
  return    0 ;
}

/* --------------------------------------------------------------------------- *
 *                 public functions: session key encapsulation                 *
 * --------------------------------------------------------------------------- */

char *
peks_wrap_session_key
  (const char    *s,
   unsigned     len,
   const char *type)
{
  char *t, *sk ;

  POINT_OF_RANDOM_STACK (7);

  if (len == 0)
    len = strlen (s) ;
  if (type == 0)
    type = "." ;

  /* make reciever session key as
   *
   *	key: <base64-key-bits> <cipher-type>
   */
  sk = bin2base64 (s, len) ;
  t  = XMALLOC (8 + strlen (sk) + strlen (type));

  POINT_OF_RANDOM_VAR (sk);

  sprintf (t, "key: %s %s", sk, type) ;
  XFREE (sk) ;

  return t;
}


char *
peks_unwrap_session_key
  (const char **S,
   unsigned    *N,
   const char *in)
{
  char *s, *sk, *sl;
  unsigned dummy ;

  if (N == 0)
    N = &dummy ;

  POINT_OF_RANDOM_STACK (7);

  /* must be editbale in order to use strtok () */
  s = ALLOCASTRDUP (in) ;

  /* 
   * parse the session key details:
   *
   *	key: <base64-key-bits> <cipher-type>
   */
  
  if ((sk = strtok (s, PEKS_FS)) == 0 ||
      strcmp (sk, "key:")        != 0 ||
      (sk = strtok (0, PEKS_FS)) == 0 ||
      (sl = strtok (0, PEKS_FS)) == 0) {
    DEALLOCA (s);
    errno = errnum (DECODE_KEY_ERR) ;
    return 0 ;
  } 
  /* extract the binary key */
  sk = base64toBin (sk, N) ;

  POINT_OF_RANDOM_VAR (sk);

  /* dup the cipher type */
  if (S != 0)
    *S = XSTRDUP (sl) ;
  
  DEALLOCA (s);
  return sk;
}

/* --------------------------------------------------------------------------- *
 *                 public functions: passwd digest                             *
 * --------------------------------------------------------------------------- */

char *
peks_digest
  (const char *dgst,
   const char *txt1,
   unsigned    len1,
   const char *txt2,
   unsigned    len2)
{
  char *s, *t;
  frame_desc *md ;

  /* get message digest method */
  if (txt1 == 0 || txt2 == 0 ||
      (md = create_frame (find_frame_class (dgst), 0)) == 0) {
    errno = errnum (FUNCTION_ARG_ERROR) ;
    return 0;
  }
  
  if (len1 == 0) len1 = strlen (txt1) ;
  if (len2 == 0) len2 = strlen (txt2) ;

  /* make a hash */
  XCRCFIRST (md, txt1, len1) ;
  XCRCNEXT  (md,  " ",    1) ;
  XCRCNEXT  (md, txt2, len2) ;

  s = bin2base64 (XCRCRESULT0 (md), md->mdlen) ;
  destroy_frame (md) ;

  sprintf (t = XMALLOC (strlen (dgst) + 2 + strlen (s) + 1),
	   "%s: %s", dgst, s);
  XFREE (s);

  return t;
}
